/*
 * tcpTokenRing.c
 *
 *  Created on: 12.10.2020
 *      Author: kaige
 */

#include <stdio.h>
#include "ucos_ii.h"
#include "ucos_bsp.h"
#include "xipipsu.h"
#include "xscugic.h"

#include <ucos_int.h>

#include "lwip/tcpip.h"
#include "lwip/dhcp.h"

#include "lwip/init.h"
#include "lwip/inet.h"
#include "lwip/opt.h"
#include "lwip/ip_addr.h"
#include "lwipopts.h"
#include "lwip/sockets.h"

#define TCP_TOKEN_RING_TASK_STACK_SIZE 2048
#define TCP_TOKEN_RING_TASK_PRIO 40
#define TCP_LARGE_OBJECT_SIZE	1024*1024
#define TCP_REQUEST_SIZE		64*1024

static CPU_STK tcpTokenRingTaskStk[TCP_TOKEN_RING_TASK_STACK_SIZE];

#define SA struct sockaddr

__attribute__((weak)) void tcpReadHook(int tcpIteration, size_t bytesLeft) {return;}
__attribute__((weak)) void tcpWriteHook(int tcpIteration, size_t bytesLeft) {return;}

int tcpIteration = 0;

static char tcpLargeObject[TCP_LARGE_OBJECT_SIZE];

static int readBlock(int socket, void*buffer, size_t size, size_t blockSize){
	size_t bytesLeft = size;
	size_t bytesRead = 0;

	tcpReadHook(tcpIteration, bytesLeft);
	if(blockSize == 0)
		blockSize = size;

	do{
		if(blockSize <= bytesLeft)
			bytesRead = read(socket, buffer, blockSize);
		else
			bytesRead = read(socket, buffer, bytesLeft);
		if(bytesRead < 0)
			return -1;
		bytesLeft -= bytesRead;
		tcpReadHook(tcpIteration, bytesLeft);
	}while(bytesLeft > 0);
	return 0;
}

static int writeBlock(int socket, void*buffer, size_t size, size_t blockSize){
	size_t bytesLeft = size;
	size_t bytesWritten = 0;

	tcpWriteHook(tcpIteration, bytesLeft);
	if(blockSize == 0)
		blockSize = size;

	do{
		if(blockSize <= bytesLeft)
			bytesWritten = write(socket, buffer, blockSize);
		else
			bytesWritten = write(socket, buffer, bytesLeft);
		if(bytesWritten < 0)
			return -1;
		bytesLeft -= bytesWritten;
		tcpWriteHook(tcpIteration, bytesLeft);
	}while(bytesLeft > 0);
	return 0;
}

/*****************************************************************/
/*
 * Forward TCP packets to realize "Token Passing Ring"
 *
 * Board is Slave
 *
 * Structure:
 * Device: 		0					  1					2				3
 * 			Master(Host1)     Slave(Board1)     Slave(Board2)		Slave(Host3)
 * 				|--------------------->|---------------->|--------------->|
 * 				|<--------------------------------------------------------|
 *
 */
int tcpTokenRing(int master, uint16_t port_in, uint16_t port_out, char *address_out, void *buffer, size_t size, size_t blockSize){
    int sockfd, connfd, peerfd, len;
    int bytesLeft, bytes;
    struct sockaddr_in servaddr, client, peer;


    // assign IP, PORT
    bzero(&servaddr, sizeof(servaddr));
    servaddr.sin_family = AF_INET;
    servaddr.sin_addr.s_addr = htonl(INADDR_ANY);
    servaddr.sin_port = htons(port_in);

    bzero(&peer, sizeof(peer));
    peer.sin_family = AF_INET;
    peer.sin_addr.s_addr = inet_addr(address_out);
    peer.sin_port = htons(port_out);


    // socket create and verification
    sockfd = socket(AF_INET, SOCK_STREAM, 0);
    peerfd = socket(AF_INET, SOCK_STREAM, 0);
    if (sockfd == -1 || peerfd == -1) {
       return -1;
    }


    // Binding newly created socket to given IP and verification
    if ((bind(sockfd, (SA*)&servaddr, sizeof(servaddr))) != 0) {
        return -2;
    }

    // Now server is ready to listen and verification
    if ((listen(sockfd, 5)) != 0) {
        return -3;
    }

//    int tcp_maxseg;
//    socklen_t tcp_maxseg_len = sizeof(tcp_maxseg);
//    tcp_maxseg = 1000;
//    if(setsockopt(sockfd, IPPROTO_TCP, TCP_MAXSEG, &tcp_maxseg, tcp_maxseg_len) < 0 ) {
//           return -5;
//    }

    len = sizeof(client);

    if(master == 0){
        // Accept the data packet from client and verification
        connfd = accept(sockfd, (SA*)&client, &len);
        if (connfd < 0) {
            return -6;
        }

        // connect the client socket to server socket
        while (connect(peerfd, (SA*)&peer, sizeof(peer)) != 0) {
        	OSTimeDly(OS_TICKS_PER_SEC);
        }

        // infinite loop for chat
        for (;;) {
        	if(readBlock(peerfd, buffer, size, blockSize) < 0)
        		return -7;
        	if(writeBlock(connfd, buffer, size, blockSize) < 0)
        		return -8;
        	tcpIteration++;
        }
    } else {
        // connect the client socket to server socket
        while (connect(peerfd, (SA*)&peer, sizeof(peer)) != 0) {
            OSTimeDly(OS_TICKS_PER_SEC);
        }
        // Accept the data packet from client and verification
        connfd = accept(sockfd, (SA*)&client, &len);
        if (connfd < 0) {
            return -6;
        }

        // infinite loop for chat
        for (;;) {
        	if(writeBlock(connfd, buffer, size, blockSize) < 0)
				return -8;
        	if(readBlock(peerfd, buffer, size, blockSize) < 0)
				return -7;
        	tcpIteration++;
        }
    }

    // After chatting close the socket
    close(peerfd);
    close(sockfd);
}

void tcpTokenRingTask(void* p_arg){

#if BOARD == 1
	tcpTokenRing(0, 1000, 1001, "192.168.178.11", tcpLargeObject, TCP_LARGE_OBJECT_SIZE, TCP_REQUEST_SIZE);
#elif BOARD == 2
	tcpTokenRing(0, 1001, 1002, "192.168.178.20", tcpLargeObject, TCP_LARGE_OBJECT_SIZE, TCP_REQUEST_SIZE);
#else
#error("Please specify board")
#endif
	OSTaskSuspend(OS_PRIO_SELF);
}

void tcpTokenRingInit(void){
	OSTaskCreateExt(tcpTokenRingTask,
						NULL,
						&tcpTokenRingTaskStk[TCP_TOKEN_RING_TASK_STACK_SIZE - 1],
						TCP_TOKEN_RING_TASK_PRIO,
						TCP_TOKEN_RING_TASK_PRIO,
						&tcpTokenRingTaskStk[0],
						TCP_TOKEN_RING_TASK_STACK_SIZE,
						DEF_NULL,
						(OS_TASK_OPT_STK_CLR | OS_TASK_OPT_STK_CHK));
	OSTaskNameSet(TCP_TOKEN_RING_TASK_PRIO, (INT8U *) "tcpTokenRingTask", NULL);
}



