/*
 * imw_types.h
 *
 *  Created on: May 5, 2020
 *      Author: kaige
 */

#ifndef _IMW_TYPES_H_
#define _IMW_TYPES_H_

#include "imw_config.h"
#include "stdint.h"
#include "imw_header.h"

/**
 * \brief Error Types for IMW functions
 */
typedef enum{
	IMW_ERR_NONE	= 	0,		/**< \brief No error occured 					*/
	IMW_ERR_ARG,				/**< \brief Argument Check error, Arg==NULL 	*/
	IMW_ERR_SOCKET,				/**< \brief Error during socket handling 		*/
	IMW_ERR_MEM,				/**< \brief Error during memory allocation 		*/
	IMW_ERR_OBJECT_ID,			/**< \brief Unexpected Object ID 				*/
	IMW_ERR_SEQ_NR,				/**< \brief Unexpected Sequence Number 			*/
	IMW_ERR_SYNC,				/**< \brief Error during synchronization        */
	IMW_ERR_OS,					/**< \brief Any OS Related Error				*/
}IMW_ERR_t;

typedef struct{
	int		socket;
	void	*rxConfig;
}IMW_RX_TO_SOCKET_MAPPING;

typedef struct{
	int				proxySocket;
	IMW_RX_TO_SOCKET_MAPPING socketMapping[IMW_MAX_READERS_PER_TASK];
	int				numberOfConfigs;
}IMW_RX_TASK_INTERNAL_t;

typedef IMW_ERR_t (*IMW_MESSAGE_HANDLER)(void* p_arg);

typedef struct{
	IMW_RX_TASK_INTERNAL_t *internal;
	int			ready;
	void *taskStack;
	uint32_t taskStackSize;
	uint8_t taskPrio;
	void *rxBuffer;
	IMW_MESSAGE_HANDLER	handleSubmessageSync;
	IMW_MESSAGE_HANDLER handleSubmessageData;
	IMW_MESSAGE_HANDLER	handleSubmessageBitmap;
}IMW_RX_TASK_CONFIG_t;

/**
 * \brief Definition of a bitmap for retransmissions
 */
typedef struct __attribute__((packed)){
	uint16_t	seqNr;							/**< \brief Object sequence number for that bitmap 		*/
	uint16_t	unused;							/**< \brief Unused, to align bitmap to word boundary 	*/
	uint32_t	bitmap[IMW_BITMAP_LENGTH/sizeof(uint32_t)];		/**< \brief Bitmap field */
}IMW_BITMAP_t;

typedef struct{
	uint16_t	objectId;
	uint32_t	size;
	uint8_t		historyCount; 					/**< \brief Size of history cache */
	void		*baseAddr;
}IMW_OBJECT_t;

typedef IMW_ERR_t (*IMW_DATA_FKT)(void* p_arg);

typedef struct{
	uint16_t		currentObjectSeqNr;
}IMW_DATA_WRITER_INTERNAL_t;

typedef struct{
	IMW_OBJECT_t	*object;
	int16_t		seqNrOffset;
	void*			taskStack;
	uint32_t		taskStackSize;
	uint8_t			taskPrio;
	uint8_t			taskId;
	IMW_DATA_FKT function;
	IMW_DATA_WRITER_INTERNAL_t*	internal;
}IMW_DATA_WRITER_CONFIG_t;

typedef struct{
	uint16_t		currentSeqNr;
}IMW_DATA_READER_INTERNAL_t;

typedef struct{
//	IMW_OBJECT_t	*object;
//	int16_t		seqNrOffset;
	void*			taskStack;
	uint32_t		taskStackSize;
	uint8_t			taskPrio;
	uint8_t			taskId;
	IMW_DATA_FKT function;
	IMW_DATA_READER_INTERNAL_t *internal;
}IMW_DATA_READER_CONFIG_t;

typedef struct{
	uint16_t		currentObjectSeqNr;
	uint16_t		numberOfFragments;
	int 			socket;
	IMW_BITMAP_t* 	currentBitmap;
	uint32_t		startTS;
}IMW_WRITER_INTERNAL_t;

typedef struct{
	IMW_OBJECT_t	*object;
	int16_t			seqNrOffset;
	uint16_t		fragmentSize;
	uint8_t			padding;
	void*			target;
	int				localPort;
	void*			txBuffer;
	void*			rxBuffer;
	uint16_t		period;
	IMW_BITMAP_t	*bitmap;
	IMW_WRITER_INTERNAL_t *internal;
	void*			taskStack;
	uint32_t		taskStackSize;
	uint8_t			taskPrio;
	uint8_t			taskId;
	IMW_RX_TASK_CONFIG_t *rxTaskConfig;
}IMW_WRITER_CONFIG_t;

typedef struct{
	uint16_t		currentObjectSeqNr;
	uint16_t		numberOfFragments;
	int 			socket;
	IMW_BITMAP_t*	bitmap;
	uint8_t			syncReceived;
	uint8_t			bitmapCheckNr;
#if IMW_READER_STATS == 1
	uint32_t		objectsSuccessfull;
	uint32_t		objectsMissing;
	uint32_t		bitmapsSend;
#endif
}IMW_READER_INTERNAL_t;

typedef struct{
	IMW_OBJECT_t	*object;
	int16_t			seqNrOffset;
	uint16_t		fragmentSize;
	uint8_t			padding;
	uint8_t			priority;
	uint8_t			maxRetry;
	void*			target;
	int				localPort;
	void*			rxBuffer;
	void*			txBuffer;
	IMW_READER_INTERNAL_t *internal;
}IMW_READER_CONFIG_t;

#define IMW_DECLARE_HOOK(name,argType)		__attribute__((weak)) void name(argType *p_arg) { (void) p_arg; return; }

#define IMW_BITMAP_SET_BIT(bm, fragNr)	 bm->bitmap[(fragNr) / 32] |= (1 << ((fragNr)%32));

#define IMW_ARG_NULL_CHECK(pointer) if((void*)(pointer)==(void*)(0)){return IMW_ERR_ARG;}

#define IMW_CEIL_DIV(a, b) (((a) + (b) - 1)/(b))

#endif /* _IMW_TYPES_H_ */
