/*
 * imw_net.c
 *
 *  Created on: May 15, 2020
 *      Author: kaige
 */

#include "imw_net.h"
#include "lwip/sockets.h"

static uint16_t	_imw_global_tx_seqNr = 0;

static void _imw_net_rx_task(void *p_arg);

#if defined(IDA_LWIP) && !defined(IDA_MW_DISABLE_RX_OPT)
static volatile IMW_ERR_t _imw_net_readHeader(int socket, IMW_SUBMESSAGE_TYPE *type){
	IMW_HEADER_t header;
	/* Receive from socket */
	recv(socket,(void*)(&header),sizeof(IMW_HEADER_t),0);

	*type = header.submessageType;
}
#else
static volatile IMW_ERR_t _imw_net_readHeader(void* buffer, IMW_SUBMESSAGE_TYPE *type){
	IMW_HEADER_t *header = (IMW_HEADER_t*)buffer;
	*type = header->submessageType;
}
#endif

IMW_ERR_t imw_net_createHeader(void* buffer, IMW_SUBMESSAGE_TYPE type){
	IMW_HEADER_t *header = (IMW_HEADER_t*)buffer;
	header->seqNr = _imw_global_tx_seqNr++;
	header->submessageType = type;
}

static uint8_t _imw_null_buffer[1500];
IMW_ERR_t imw_net_flush_socket(int socket){
	recv(socket,(void*)_imw_null_buffer,sizeof(_imw_null_buffer),0);
}

IMW_ERR_t imw_net_rx_register_handler(IMW_RX_TASK_CONFIG_t *rxTaskConfig, int socket, void* rxConfig){
	IMW_RX_TASK_INTERNAL_t *internal;
	int sockfd;
	IMW_ARG_NULL_CHECK(rxTaskConfig);
	IMW_ARG_NULL_CHECK(rxConfig);
	internal = rxTaskConfig->internal;
	IMW_ARG_NULL_CHECK(internal);

	sockfd = socket;

	if(internal->numberOfConfigs == IMW_MAX_READERS_PER_TASK)
		return IMW_ERR_MEM;

	int configNr = internal->numberOfConfigs;
	internal->socketMapping[configNr].socket = sockfd;
	internal->socketMapping[configNr].rxConfig = rxConfig;
#ifdef IDA_LWIP
	sendto(internal->proxySocket, &sockfd, sizeof(int), 0, NULL, 0);
#endif
	internal->numberOfConfigs++;

	return IMW_ERR_NONE;
}



IMW_ERR_t imw_net_rx_task_init(IMW_RX_TASK_CONFIG_t *taskConfig){
	IMW_RX_TASK_INTERNAL_t *internal;

	/* Argument checking */
	IMW_ARG_NULL_CHECK(taskConfig);
	IMW_ARG_NULL_CHECK(taskConfig->internal);

	internal = taskConfig->internal;

#ifdef IDA_LWIP
	/* Create a proxy socket */
	internal->proxySocket = socket(PF_INET, SOCK_PROXY, 0);
	if(internal->proxySocket < 0)
		return IMW_ERR_SOCKET;
#endif

	for(uint16_t i = 0; i < IMW_MAX_READERS_PER_TASK; i++){
		internal->socketMapping[i].rxConfig = (void*)0;
		internal->socketMapping[i].socket = -1;
	}
	internal->numberOfConfigs = 0;

	OSTaskCreateExt( _imw_net_rx_task,
				(void*)taskConfig,
				&((OS_STK*)taskConfig->taskStack)[taskConfig->taskStackSize - 1],
				taskConfig->taskPrio,
				taskConfig->taskPrio,
				(OS_STK*)taskConfig->taskStack,
				taskConfig->taskStackSize,
				 DEF_NULL,
				(OS_TASK_OPT_STK_CLR | OS_TASK_OPT_STK_CHK));

	OSTaskNameSet(taskConfig->taskPrio,(INT8U *)"imwRXTask",NULL);

//	while(taskConfig->ready == 0){
//		OSTimeDly(OS_TICKS_PER_SEC/100);
//	}

	return IMW_ERR_NONE;
}

static void _imw_net_rx_task(void *p_arg){
	IMW_RX_TASK_CONFIG_t *taskConfig = (IMW_RX_TASK_CONFIG_t*)p_arg;
	int rxReady_fd = -1;				/* RX Socket Ready for read */
	IMW_SUBMESSAGE_TYPE submessageType;
	void* rxConfig;
	IMW_RX_TASK_INTERNAL_t *internal;
#ifndef IDA_LWIP
	fd_set rset;
	int maxfdp1, nready;
#endif

	/* Argument checking */
	if(taskConfig == (IMW_RX_TASK_CONFIG_t*)0)
		return;

	internal = taskConfig->internal;
	if(internal == (IMW_RX_TASK_INTERNAL_t*)0)
		return;

	taskConfig->ready = 1;

	while(1){
#ifdef IDA_LWIP
		if(recv(internal->proxySocket, &rxReady_fd, sizeof(rxReady_fd),0) != -1){
			for(uint16_t i = 0; i < internal->numberOfConfigs; i++)
				if(internal->socketMapping[i].socket == rxReady_fd)
					rxConfig = internal->socketMapping[i].rxConfig;
#else
		FD_ZERO(&rset);
		maxfdp1 = 0;
		for(uint16_t i = 0; i < internal->numberOfConfigs; i++){
			if(internal->socketMapping[i].socket > maxfdp1)
				maxfdp1 = internal->socketMapping[i].socket;
			FD_SET(internal->socketMapping[i].socket, &rset);
		}
		maxfdp1 += 1;
		nready = select(maxfdp1, &rset, NULL, NULL, NULL);
		for(uint16_t i = 0; i < internal->numberOfConfigs; i++){
			if(FD_ISSET(internal->socketMapping[i].socket, &rset)){
				rxConfig = internal->socketMapping[i].rxConfig;
				rxReady_fd = internal->socketMapping[i].socket;
#endif

#if defined(IDA_LWIP) && !defined(IDA_MW_DISABLE_RX_OPT)
			_imw_net_readHeader(rxReady_fd,&submessageType);
#else
			recv(rxReady_fd,taskConfig->rxBuffer,1500,0);
			_imw_net_readHeader(taskConfig->rxBuffer,&submessageType);
#endif
			switch(submessageType){
//			case IMW_SUBMESSAGE_SYNC:
//				if(taskConfig->handleSubmessageSync != (void*)0)
//					taskConfig->handleSubmessageSync(rxConfig);
				break;
			case IMW_SUBMESSAGE_DATA:
				if(taskConfig->handleSubmessageData != (void*)0)
					taskConfig->handleSubmessageData(rxConfig);
				break;
			case IMW_SUBMESSAGE_BITMAP:
				if(taskConfig->handleSubmessageBitmap != (void*)0)
					taskConfig->handleSubmessageBitmap(rxConfig);
				break;
			default:
				break;
			}
		}
#ifndef IDA_LWIP
		}
#endif
	}

}
