/*
 * imw_reader.c
 *
 *  Created on: May 6, 2020
 *      Author: kaige
 */

#include "imw_reader.h"
#include "lwip/sockets.h"

IMW_DECLARE_HOOK(imw_reader_sumbessageData_start_hook, IMW_SUBMESSAGE_DATA_HEADER_t);
IMW_DECLARE_HOOK(imw_reader_sumbessageData_end_hook, IMW_SUBMESSAGE_DATA_HEADER_t);
IMW_DECLARE_HOOK(imw_reader_bitmapCheckFailed_hook, IMW_READER_CONFIG_t);

LWIP_MEMPOOL_DECLARE(IMW_BITMAP_POOL, 2, sizeof(IMW_BITMAP_t), "tx Pbuf_Pool");

IMW_ERR_t imw_reader_Init(void){
	LWIP_MEMPOOL_INIT(IMW_BITMAP_POOL);
}

static IMW_ERR_t _imw_allocBitmap(IMW_BITMAP_t **bitmapStorage, uint16_t seqNr){
	IMW_BITMAP_t * bitmap;
	bitmap = (IMW_BITMAP_t*)LWIP_MEMPOOL_ALLOC(IMW_BITMAP_POOL);
	if(bitmap == (IMW_BITMAP_t*)0){
		return IMW_ERR_MEM;
	}
	bitmap->seqNr = seqNr;
	memset(bitmap->bitmap,0,IMW_BITMAP_LENGTH/sizeof(uint32_t));
	*bitmapStorage = bitmap;
	return IMW_ERR_NONE;
}

//IMW_ERR_t imw_reader_handleSubmessageSync(void* readerCfg){
//	IMW_READER_CONFIG_t *readerConfig = (IMW_READER_CONFIG_t*)readerCfg;
//	IMW_READER_INTERNAL_t *internal;
//	IMW_OBJECT_t *object;
//	IMW_SUBMESSAGE_SYNC_HEADER_t header;
//	uint8_t discardPadding[8];
//
//	IMW_ARG_NULL_CHECK(readerConfig);
//	IMW_ARG_NULL_CHECK(readerConfig->internal);
//	IMW_ARG_NULL_CHECK(readerConfig->object);
//
//	internal = readerConfig->internal;
//	object = readerConfig->object;
//
//	/* Receive from socket */
//	recv(internal->socket,(void*)(&header),sizeof(IMW_SUBMESSAGE_SYNC_HEADER_t),0);
//
//	/* Check if the object id is ok */
//	if(header.objectId != object->objectId)
//		return IMW_ERR_OBJECT_ID;
//
//	internal->currentObjectSeqNr = header.objectSequenceNr;
//	internal->syncReceived = 1;
//	return IMW_ERR_NONE;
//}

IMW_ERR_t imw_reader_handleSubmessageData(void* readerCfg){
	IMW_READER_CONFIG_t* readerConfig = (IMW_READER_CONFIG_t*)readerCfg;
	IMW_READER_INTERNAL_t *internal;
	IMW_OBJECT_t *object;
	IMW_SUBMESSAGE_DATA_HEADER_t header;
	uint8_t discardPadding[8];
	uint32_t debugData[2];

	IMW_ARG_NULL_CHECK(readerConfig);
	IMW_ARG_NULL_CHECK(readerConfig->internal);
	IMW_ARG_NULL_CHECK(readerConfig->object);

	internal = readerConfig->internal;
	object = readerConfig->object;

	/* Receive from socket */
#if defined(IDA_LWIP) && !defined(IDA_MW_DISABLE_RX_OPT)
	recv(internal->socket,(void*)(&header),sizeof(IMW_SUBMESSAGE_DATA_HEADER_t),0);
#else
	memcpy((void*)&header,(void*)(readerConfig->rxBuffer + sizeof(IMW_HEADER_t)), sizeof(IMW_SUBMESSAGE_DATA_HEADER_t));
#endif

	imw_reader_sumbessageData_start_hook(&header);

#if defined(IDA_LWIP) && !defined(IDA_MW_DISABLE_RX_OPT)
	recv(internal->socket,(void*)discardPadding,header.padding,0);
#endif

	/* Check if the object id is ok */
	if(header.objectId != object->objectId){
		imw_net_flush_socket(internal->socket);
		return IMW_ERR_OBJECT_ID;
	}

//	/* Discard old packets */
//	if(header.objectSequenceNr != internal->currentObjectSeqNr){
//		imw_net_flush_socket(internal->socket);
//		return IMW_ERR_SEQ_NR;
//	}

	/* Check if bitmap is up-to-date and set corresponding bit */
	IMW_BITMAP_t *bitmap = internal->bitmap;
	if(bitmap->seqNr != header.objectSequenceNr){
		imw_reader_check_bitmap(readerConfig);
		/* We need to create a new bitmap for the new sequence number */
		LWIP_MEMPOOL_FREE(IMW_BITMAP_POOL,(void*)bitmap);
		if(_imw_allocBitmap(&internal->bitmap,header.objectSequenceNr) != IMW_ERR_NONE){
			imw_net_flush_socket(internal->socket);
			return IMW_ERR_MEM;
		}
	}
	IMW_BITMAP_SET_BIT(bitmap, header.fragmentId);

	/* Calculate base address of object-instance in history cache */
	void* base = ((header.objectSequenceNr % object->historyCount) * object->size) + object->baseAddr;
	/* Add offset of current fragment */
	base += header.fragmentId * readerConfig->fragmentSize;

#if defined(IDA_LWIP) && !defined(IDA_MW_DISABLE_RX_OPT)
	recv(internal->socket,base,header.fragmentLength,0);
#else
	memcpy(base,(void*)((readerConfig->rxBuffer) + sizeof(IMW_HEADER_t) + sizeof(IMW_SUBMESSAGE_DATA_HEADER_t) + header.padding),header.fragmentLength);
#endif
	imw_reader_sumbessageData_end_hook(&header);
	return IMW_ERR_NONE;
}



int imw_reader_create(IMW_READER_CONFIG_t *config){
	int sockfd = -1;
	struct sockaddr_in local;
	IMW_READER_INTERNAL_t *internal;
	IMW_ARG_NULL_CHECK(config);

	internal = config->internal;

	sockfd = socket(AF_INET, SOCK_DGRAM, 0);
	if (sockfd == -1) {
		return -1;
	}

	memset(&local, 0, sizeof(struct sockaddr_in));
	local.sin_family = AF_INET;
	local.sin_addr.s_addr = htonl(INADDR_ANY);
	local.sin_port = htons(config->localPort);
	if(bind(sockfd, (struct sockaddr *)&local, sizeof(struct sockaddr_in)) == -1){
		close(sockfd);
		return -1;
	}

#if defined(IDA_LWIP) && !defined(IDA_MW_DISABLE_RX_OPT)
	ida_lwip_set_socket_prio(sockfd,config->priority);
#endif

	internal->socket = sockfd;
	internal->currentObjectSeqNr = config->seqNrOffset;
	internal->objectsMissing = 0;
	internal->objectsSuccessfull = 0;
	internal->bitmapsSend = 0;
	internal->bitmapCheckNr = 0;

	if(_imw_allocBitmap(&internal->bitmap, 0) != IMW_ERR_NONE){
		close(sockfd);
		return -1;
	}

	internal->numberOfFragments = config->object->size / config->fragmentSize;
	if(config->object->size % config->fragmentSize != 0)
		internal->numberOfFragments++;

	return sockfd;
}

IMW_ERR_t imw_reader_send_bitmap(IMW_READER_CONFIG_t *config){
	IMW_READER_INTERNAL_t *internal;
	IMW_SUBMESSAGE_BITMAP_HEADER_t	*bmHeader;

	internal = config->internal;

	void* buffer = config->txBuffer;
	imw_net_createHeader(buffer,IMW_SUBMESSAGE_BITMAP);
	buffer += sizeof(IMW_HEADER_t);

	/* Bitmap size in bytes */
	uint16_t bitmapSize = ((internal->numberOfFragments+1)/32)*sizeof(uint32_t);

	bmHeader = (IMW_SUBMESSAGE_BITMAP_HEADER_t*)buffer;
	bmHeader->objectId = config->object->objectId;
	bmHeader->objectSequenceNr = internal->currentObjectSeqNr;
	bmHeader->bitmapSize = bitmapSize;
	buffer += sizeof(IMW_SUBMESSAGE_BITMAP_HEADER_t);

	memcpy(buffer,(void*)&(internal->bitmap->bitmap[0]),bmHeader->bitmapSize * sizeof(uint32_t));

	sendto(internal->socket, config->txBuffer, sizeof(IMW_HEADER_t) + sizeof(IMW_SUBMESSAGE_BITMAP_HEADER_t) + bitmapSize,0, (const struct sockaddr*)config->target, sizeof(struct sockaddr_in));
}

IMW_ERR_t imw_reader_check_bitmap(IMW_READER_CONFIG_t *config){
	IMW_READER_INTERNAL_t *internal;
	IMW_BITMAP_t *bm;
	IMW_ARG_NULL_CHECK(config);

	internal = config->internal;

	bm = internal->bitmap;

	int remaining = internal->numberOfFragments;
	int rxSuccessfull = 1;
	int i = 0;
	while(remaining > 0){
		if(remaining > 32){
			if(~(bm->bitmap[i]) != 0){
				rxSuccessfull = 0;
				break;
			}
		} else {
			uint32_t bits = (~(bm->bitmap[i]) & (((uint32_t)1 << remaining) - 1));
			if(bits != 0){
				rxSuccessfull = 0;
			}
		}
		remaining -= 32;
		i++;
	}

	internal->bitmapCheckNr = (internal->bitmapCheckNr + 1) % config->maxRetry;
	if(internal->bitmapCheckNr == 0){
		/* Final check, no retransmission possible */
		if(!rxSuccessfull){
			imw_reader_bitmapCheckFailed_hook(config);
			internal->objectsMissing++;
		} else {
			internal->objectsSuccessfull++;
		}
	} else {
		/* Check if all fragments are present, otherwise request retransmission of missing fragements */
		if(!rxSuccessfull){
			imw_reader_send_bitmap(config);
			internal->bitmapsSend++;
		}
	}

	return IMW_ERR_NONE;
}
