/*
 * imw_writer.c
 *
 *  Created on: May 5, 2020
 *      Author: kaige
 */

#include "imw_writer.h"
#include "lwip/sockets.h"

static void _imwWriterTask(void *parg);

IMW_DECLARE_HOOK(imw_writer_start_hook, IMW_WRITER_CONFIG_t);
IMW_DECLARE_HOOK(imw_writer_end_hook, IMW_WRITER_CONFIG_t);
__attribute__((weak)) void imw_writer_send_fragment_hook(uint16_t seqNr, int fragment) { (void) seqNr; return; }

IMW_ERR_t imw_writer_create(IMW_WRITER_CONFIG_t *config){
	uint8_t err = 0;
	int sockfd = -1;
	struct sockaddr_in local;

	sockfd = socket(AF_INET, SOCK_DGRAM, 0);
	if (sockfd == -1) {
		return IMW_ERR_SOCKET;
	}

	memset(&local, 0, sizeof(struct sockaddr_in));
	local.sin_family = AF_INET;
	local.sin_addr.s_addr = htonl(INADDR_ANY);
	local.sin_port = htons(config->localPort);
	if(bind(sockfd, (struct sockaddr *)&local, sizeof(struct sockaddr_in)) == -1){
		close(sockfd);
		return IMW_ERR_SOCKET;
	}

	config->internal->socket = sockfd;
	config->internal->currentObjectSeqNr = 0;//config->seqNrOffset;
	config->internal->currentBitmap = (IMW_BITMAP_t*)0;

	config->internal->numberOfFragments = IMW_CEIL_DIV(config->object->size, config->fragmentSize);

	err = OSTaskCreateExt( _imwWriterTask,
				(void*)config,
				&((OS_STK*)(config->taskStack))[config->taskStackSize -1],
				config->taskPrio,
				config->taskId,
				(OS_STK*)(config->taskStack),
				config->taskStackSize,
				 DEF_NULL,
				(OS_TASK_OPT_STK_CLR | OS_TASK_OPT_STK_CHK));
	if(err != OS_ERR_NONE){
		close(sockfd);
		return IMW_ERR_OS;
	}

	OSTaskNameSet(config->taskPrio,(INT8U *)"imwWriterTask",NULL);

	return IMW_ERR_NONE;
}

static IMW_ERR_t _imw_writer_send_fragment(IMW_WRITER_CONFIG_t *config, int socket, uint16_t seqNr, void* base, int fragment){
	void* txBuffer;
	uint32_t bytesToTransfer = 0;
	IMW_OBJECT_t *object;
	IMW_WRITER_INTERNAL_t *internal;

	IMW_ARG_NULL_CHECK(config);
	IMW_ARG_NULL_CHECK(config->object);
	IMW_ARG_NULL_CHECK(config->internal);

	imw_writer_send_fragment_hook(seqNr, fragment);

	/*Reset txBuffer */
	txBuffer = config->txBuffer;
	object = config->object;

	/* Calculate size of packet */
	bytesToTransfer = config->fragmentSize;
	if(object->size % config->fragmentSize != 0){
		if(fragment == config->internal->numberOfFragments - 1){
			bytesToTransfer = object->size % config->fragmentSize;
		}
	}

	/* Create message Header */
	imw_net_createHeader(txBuffer, IMW_SUBMESSAGE_DATA);
	txBuffer += sizeof(IMW_HEADER_t);

	/* Write data submessage header */
	IMW_SUBMESSAGE_DATA_HEADER_t *header = (IMW_SUBMESSAGE_DATA_HEADER_t*)txBuffer;
	header->objectId = object->objectId;
	header->objectSequenceNr = seqNr;
	header->fragmentId = fragment;
	header->fragmentLength = bytesToTransfer;
	header->padding = config->padding;

#if defined(IDA_LWIP) && !defined(IDA_MW_DISABLE_TX_OPT)
	struct msghdr message;
	struct iovec msg_iov[2];

	msg_iov[0].iov_base = config->txBuffer;
	msg_iov[0].iov_len = sizeof(IMW_HEADER_t) + sizeof(IMW_SUBMESSAGE_DATA_HEADER_t) + config->padding;
	msg_iov[1].iov_base = base+(fragment*config->fragmentSize);
	msg_iov[1].iov_len = bytesToTransfer;
	message.msg_iov = &msg_iov[0];
	message.msg_iovlen = 2;
	message.msg_name = (void*)config->target;

	ssize_t x = sendmsg(socket, &message, 0);
#else
	/* Add padding */
	txBuffer += sizeof(IMW_SUBMESSAGE_DATA_HEADER_t);
	txBuffer += config->padding;

	/* Copy data to buffer, needs to be replaced by scatter gather */
	memcpy(txBuffer,base+(fragment*config->fragmentSize), bytesToTransfer);

	/* Calculate total size of packet */
	bytesToTransfer += sizeof(IMW_HEADER_t) + sizeof(IMW_SUBMESSAGE_DATA_HEADER_t) + config->padding;

	/* Fire and forget */
	sendto(socket, config->txBuffer, bytesToTransfer,0, (const struct sockaddr*)config->target, sizeof(struct sockaddr_in));
#endif
}

IMW_ERR_t imw_writer_retransmit(IMW_WRITER_CONFIG_t* config){
	IMW_OBJECT_t *object;
	IMW_WRITER_INTERNAL_t *internal;
	IMW_BITMAP_t *bitmap;
	uint8_t bitmapLines = 0;
	uint16_t fragmentNr = 0;

	IMW_ARG_NULL_CHECK(config);
	IMW_ARG_NULL_CHECK(config->internal);
	IMW_ARG_NULL_CHECK(config->object);

	object = config->object;
	internal = config->internal;

	/* Check if there is something to retransmit, indicated by a bitmap */
	if(internal->currentBitmap == (IMW_BITMAP_t*)0)
		return IMW_ERR_NONE;

	bitmap = internal->currentBitmap;

	/**
	 * Calculate Base-Address of this sample based on the seqNr in the bitmap
	 * */
	void* base = (((uint32_t)(bitmap->seqNr) % object->historyCount) * object->size) + object->baseAddr;

	/* Iterate over bitmap lines and send missing parts */
	bitmapLines = (internal->numberOfFragments + 31)/32;
	for(uint32_t line = 0; line < bitmapLines; line++){
		uint32_t entry = bitmap->bitmap[line];
		if(entry != 0xFFFFFFFF){
			for(uint16_t bit = 0; bit < 32; bit++){
				if((entry & (1 << bit)) == 0){
					fragmentNr = (line * 32) + bit;
					_imw_writer_send_fragment(config,internal->socket,bitmap->seqNr,base,fragmentNr);

				}
			}
		}
	}

	/* We are done (for now), remove bitmap reference */
	internal->currentBitmap = (IMW_BITMAP_t*)0;
	return IMW_ERR_NONE;
}

/**
 * Write data to the current object entry
 */
IMW_ERR_t imw_netDeaySim(IMW_WRITER_CONFIG_t *config){
	IMW_OBJECT_t *object;
	IMW_WRITER_INTERNAL_t *internal;

	IMW_ARG_NULL_CHECK(config);
	IMW_ARG_NULL_CHECK(config->internal);

	internal = config->internal;

	uint32_t time = OSTimeGet();
	if(internal->currentObjectSeqNr == 0){
		internal->startTS = time;
	}
	uint32_t diff = (time - internal->startTS) - (internal->currentObjectSeqNr * config->period);

	uint32_t delay = rand() / (RAND_MAX / (120-diff) + 1);
	OSTimeDly(delay);
}

IMW_ERR_t imw_writer_write(IMW_WRITER_CONFIG_t* config){
	IMW_OBJECT_t *object;
	IMW_WRITER_INTERNAL_t *internal;

	IMW_ARG_NULL_CHECK(config);
	IMW_ARG_NULL_CHECK(config->object);
	IMW_ARG_NULL_CHECK(config->internal);

	object = config->object;
	internal = config->internal;

//	imw_netDeaySim(config);

	void* base = (((uint32_t)(internal->currentObjectSeqNr + config->seqNrOffset) % object->historyCount) * object->size) + object->baseAddr;

	/* Transmit every fragment */
	for(uint16_t fragment = 0; fragment < internal->numberOfFragments; fragment++){
		_imw_writer_send_fragment(config,internal->socket,internal->currentObjectSeqNr,base,fragment);
	}

	return IMW_ERR_NONE;
}

IMW_ERR_t imw_writer_handleSubmessageBitmap(void *p_arg){
	IMW_WRITER_CONFIG_t *config = (IMW_WRITER_CONFIG_t*)p_arg;
	IMW_WRITER_INTERNAL_t *internal;
	IMW_OBJECT_t *object;
	IMW_SUBMESSAGE_BITMAP_HEADER_t header;

	IMW_ARG_NULL_CHECK(config);
	IMW_ARG_NULL_CHECK(config->internal);
	IMW_ARG_NULL_CHECK(config->object);

	internal = config->internal;
	object = config->object;

	/* Receive from socket */
	recv(internal->socket,(void*)(&header),sizeof(IMW_SUBMESSAGE_BITMAP_HEADER_t),0);

	/* Check if the object id is ok */
	if(header.objectId != object->objectId){
		return IMW_ERR_OBJECT_ID;
	}

	/**
	 * Note: currentObjectSeqNr is increased at the end of the write phase,
	 * therefore we accept bitmaps only for the previous seqNr
	 */
	if(header.objectSequenceNr != internal->currentObjectSeqNr-1){
		return IMW_ERR_SEQ_NR;
	}

	internal->currentBitmap = config->bitmap;
	internal->currentBitmap->seqNr = header.objectSequenceNr;
	recv(internal->socket,(void*)internal->currentBitmap->bitmap,header.bitmapSize,0);

	return IMW_ERR_NONE;
}

extern void ILM_OS_RegisterActSem(uint8_t TaskId, OS_EVENT* actSem);

static void _imwWriterTask(void *parg){
	IMW_WRITER_CONFIG_t *config = (IMW_WRITER_CONFIG_t*)parg;

	uint8_t err;

	OS_EVENT *sem = OSSemCreate(0);
	ILM_OS_RegisterActSem(config->taskId, sem);

	imw_net_rx_register_handler(config->rxTaskConfig,config->internal->socket,(void*)&config);

	while(1){
		OSSemPend(sem,0, &err);
		imw_writer_start_hook(config);
		if(imw_writer_write(config) != IMW_ERR_NONE)
			OSTaskDel(OS_PRIO_SELF);
		imw_writer_end_hook(config);
		config->internal->currentObjectSeqNr++;
	//		OSSemPend(sem,0, &err);
	//		if(imw_writer_retransmit((IMW_WRITER_CONFIG_t*)parg) != IMW_ERR_NONE)
	//			OSTaskDel(OS_PRIO_SELF);
	}
}
