/* net.c */

#include "ptpd.h"
#include "lwip/igmp.h"
#include "lwip/sockets.h"
#include <lwip/inet.h>
#include "netif/ethernet.h"

#include "netif/xemacps_ieee1588.h"
#ifdef IDA_LWIP
#include "ida-lwip/ida_lwip_filter.h"
#endif

#define PORT 2468

#if defined(IDA_LWIP) || defined(PTP_LAYER_2)
#define PTP_EVENT_PACKET_SIZE	58
#define PTP_GENERAL_PACKET_SIZE 78
#endif

//static int socketEvent;
//static int socketGeneral;
struct sockaddr_in multicastAddr;

#if defined(PTP_LAYER_2)
sys_sem_t ptpRxSem;
#endif

#if defined(IDA_LWIP) || defined(PTP_LAYER_2)
struct pbuf* ptp_rx_pbuf;
sys_sem_t txSem;

static struct eth_addr ownMac;
/* multicast address for PTP over ethernet */
static struct eth_addr ptp_l2_dst_addr = ETH_ADDR(0x01, 0x1b, 0x19, 0x00, 0x00, 0x00);
uint8_t event_tx_buffer[PTP_EVENT_PACKET_SIZE];
uint8_t general_tx_buffer[PTP_GENERAL_PACKET_SIZE];

#if LWIP_PBUF_TIMESTAMP != 1
#error "PTP Driver requires pbuf timestamps"
#endif
#endif

/* Shut down  the UDP and network stuff */
bool netShutdown(NetPath *netPath)
{
	struct ip4_addr multicastAaddr;

	DBG("netShutdown\n");

#ifndef IDA_LWIP
	/* leave multicast group */
//	multicastAaddr.addr = netPath->multicastAddr;
	igmp_leavegroup(IP_ADDR_ANY, (struct ip4_addr*)&netPath->multicastAddr.sin_addr);
#endif

	/* Clear the network addresses. */
	memset(&netPath->multicastAddr,0,sizeof(struct sockaddr_in));
	memset(&netPath->unicastAddr,0,sizeof(struct sockaddr_in));

	/* Return a success code. */
	return TRUE;
}

/* Find interface to  be used.  uuid should be filled with MAC address of the interface.
	 Will return the IPv4 address of  the interface. */
static int32_t findIface(const octet_t *ifaceName, octet_t *uuid, NetPath *netPath)
{
	struct netif *iface;

	iface = netif_default;
	memcpy(uuid, iface->hwaddr, iface->hwaddr_len);

	return iface->ip_addr.addr;
}

/* Start  all of the UDP stuff */
bool netInit(NetPath *netPath, PtpClock *ptpClock)
{
#if defined(PTP_LAYER_2)
	sys_sem_new(&ptpRxSem, 0);
#endif

#if defined(IDA_LWIP) || defined(PTP_LAYER_2)
	ptp_rx_pbuf = (struct pbuf*)0;
	sys_sem_new(&txSem, 0);

	memcpy(ptpClock->portUuidField, netif_default->hwaddr, netif_default->hwaddr_len);
	memcpy(ownMac.addr,netif_default->hwaddr, netif_default->hwaddr_len);
#else
//	struct in_addr netAddr;
	struct ip4_addr interfaceAddr;
	char addrStr[NET_ADDRESS_LENGTH];

	DBG("netInit\n");

	/* Find a network interface */
	interfaceAddr.addr = findIface(ptpClock->rtOpts->ifaceName, ptpClock->portUuidField, netPath);
	if (!(interfaceAddr.addr))
	{
			ERROR("netInit: Failed to find interface address\n");
			return FALSE;
	}

	struct sockaddr_in sLocalAddr;

	memset(&sLocalAddr, '\0', sizeof(sLocalAddr));
	sLocalAddr.sin_family = AF_INET;
	sLocalAddr.sin_len = sizeof(sLocalAddr);
	sLocalAddr.sin_addr.s_addr = INADDR_ANY;  /* bind to all interfaces */
	sLocalAddr.sin_port = htons(PTP_EVENT_PORT);

	netPath->socketEvent = socket(AF_INET, SOCK_DGRAM, 0);
	if (netPath->socketEvent < 0) return FALSE;
	netPath->socketGeneral = socket(AF_INET, SOCK_DGRAM, 0);
	if (netPath->socketGeneral < 0) return FALSE;

	if (bind(netPath->socketEvent, (struct sockaddr *)&sLocalAddr, sizeof(sLocalAddr)) < 0) {
		close(netPath->socketEvent);
		close(netPath->socketGeneral);
		return FALSE;
	}

	sLocalAddr.sin_port = htons(PTP_GENERAL_PORT);

	if (bind(netPath->socketGeneral, (struct sockaddr *)&sLocalAddr, sizeof(sLocalAddr)) < 0) {
		close(netPath->socketGeneral);
		close(netPath->socketGeneral);
		return FALSE;
	}

	memset(&netPath->unicastAddr,0,sizeof(struct sockaddr_in));

	/* Init General multicast IP address */
	memcpy(addrStr, DEFAULT_PTP_DOMAIN_ADDRESS, NET_ADDRESS_LENGTH);
	if (!inet_aton(addrStr, &netPath->multicastAddr.sin_addr))
	{
			ERROR("netInit: failed to encode multi-cast address: %s\n", addrStr);
			close(netPath->socketEvent);
			close(netPath->socketGeneral);
			return FALSE;
	}
	netPath->multicastAddr.sin_family = AF_INET;
	netPath->multicastAddr.sin_len = sizeof(struct sockaddr_in);
	netPath->multicastAddr.sin_port = htons(0);

	/* Join multicast group (for receiving) on specified interface */
	igmp_joingroup(&interfaceAddr, (struct ip4_addr *)&netPath->multicastAddr.sin_addr);

	/* Init Peer multicast IP address */
	memcpy(addrStr, PEER_PTP_DOMAIN_ADDRESS, NET_ADDRESS_LENGTH);
	if (!inet_aton(addrStr, &netPath->peerMulticastAddr.sin_addr.s_addr))
	{
			ERROR("netInit: failed to encode peer multi-cast address: %s\n", addrStr);
			close(netPath->socketEvent);
			close(netPath->socketGeneral);
			return FALSE;
	}
	netPath->peerMulticastAddr.sin_family = AF_INET;
	netPath->peerMulticastAddr.sin_len = sizeof(struct sockaddr_in);
	netPath->peerMulticastAddr.sin_port = htons(0);

	/* Join peer multicast group (for receiving) on specified interface */
	igmp_joingroup(&interfaceAddr, (struct ip4_addr *) &netPath->peerMulticastAddr.sin_addr.s_addr);

	/* Multicast send only on specified interface. */
//	netPath->eventPcb->multicast_ip.addr = netPath->multicastAddr;
//	netPath->generalPcb->multicast_ip.addr = netPath->multicastAddr;


	/* Return a success code. */
//	return netMasterSlaveInit(NULL);
#endif
	return TRUE;
}

/* Wait for a packet  to come in on either port.  For now, there is no wait.
 * Simply check to  see if a packet is available on either port and return 1,
 *  otherwise return 0. */
int32_t netSelect(NetPath *netPath, const TimeInternal *timeout)
{
#ifdef IDA_LWIP
	if(ptp_rx_pbuf == (struct pbuf*) 0){
		ptp_rx_pbuf = ida_filter_receivePtp(timeout->nanoseconds / 1000000);
		if(ptp_rx_pbuf == (struct pbuf*) 0)
			return 0;

	}
	return 1;

#elif defined (PTP_LAYER_2)
	if(ptp_rx_pbuf == (struct pbuf*) 0){
		sys_arch_sem_wait(&ptpRxSem,timeout->nanoseconds / 1000000);
		if(ptp_rx_pbuf == (struct pbuf*)0)
			return 0;
	}
	return 1;
#else
	struct timeval tv;
	tv.tv_sec = 0;
	tv.tv_usec = timeout->nanoseconds / 1000;
	fd_set read_sds;
	int max_sd = 0;
	if(netPath->socketEvent > netPath->socketGeneral)
		max_sd = netPath->socketEvent + 1;
	else if(netPath->socketGeneral > netPath->socketEvent)
		max_sd = netPath->socketGeneral + 1;
	FD_ZERO(&read_sds);
	FD_SET(netPath->socketEvent, &read_sds);
	FD_SET(netPath->socketGeneral, &read_sds);
	if (select(max_sd, &read_sds, NULL, NULL, &tv) <= 0)
		return 0;
	return 1;
#endif
}

/* Delete all waiting packets in event queue. */
void netEmptyEventQ(NetPath *netPath)
{
 (void)netPath;
}

ssize_t netRecvEvent(NetPath *netPath, octet_t *buf, TimeInternal *time)
{
#if defined(IDA_LWIP) || defined(PTP_LAYER_2)
	if(ptp_rx_pbuf) {
		ssize_t len = ptp_rx_pbuf->len;
		memcpy(buf,ptp_rx_pbuf->payload,len);
		time->seconds = ptp_rx_pbuf->ts_sec;
		time->nanoseconds = ptp_rx_pbuf->ts_nsec;
		pbuf_free(ptp_rx_pbuf);
		ptp_rx_pbuf = NULL;

		return len;
	}
#else
	err_t res;
	res = recv(netPath->socketEvent, buf,PACKET_SIZE,MSG_DONTWAIT);
	if(res > 0)
	{
		ETH_PTP_GetTimestamp(&time->seconds, &time->nanoseconds, TRUE);
		return res;
	}
#endif
	return 0;
}

ssize_t netRecvGeneral(NetPath *netPath, octet_t *buf, TimeInternal *time)
{
#ifdef IDA_LWIP
#else
	err_t res;
	res = recv(netPath->socketGeneral, buf,PACKET_SIZE,MSG_DONTWAIT);
	if(res > 0)
	{
		return res;
	}
#endif
	return 0;
}

static ssize_t socketSend(int socket, void *data, int16_t  length, TimeInternal *time, struct sockaddr_in * receiver)
{
#ifdef IDA_LWIP
	// TODO
	return 0;
#else
	ssize_t bytes_sent;
	struct sockaddr_in local_sin;
	socklen_t local_sinlen = sizeof(local_sin);
	getsockname(socket, (struct sockaddr*)&local_sin, &local_sinlen);
	receiver->sin_port = local_sin.sin_port;
	bytes_sent = sendto(socket, data, length, 0 ,(struct sockaddr *) receiver, sizeof(*receiver));

	if(bytes_sent < 0) {
	//		lwip_close(*sock);
		return -1;
	}

	if (time != NULL) {
		getTime(time);
	}

	return bytes_sent;
#endif
}

ssize_t netSendEvent(NetPath *netPath, const octet_t *buf, int16_t  length, TimeInternal *time)
{
#if defined(IDA_LWIP)
	struct eth_hdr *ethhdr = (struct eth_hdr *)&event_tx_buffer[0];
	ethhdr->type = lwip_htons(ETHTYPE_PTP);
	SMEMCPY(&ethhdr->dest, &ptp_l2_dst_addr, ETH_HWADDR_LEN);
	SMEMCPY(&ethhdr->src,  ownMac.addr, ETH_HWADDR_LEN);

	memcpy(&event_tx_buffer[SIZEOF_ETH_HDR],buf,length);

	size_t data_size = ida_lwip_send_raw((void*)event_tx_buffer, length+SIZEOF_ETH_HDR, &txSem);

	/* save timestamp of PTP event message */
	ETH_PTP_GetTimestamp(&time->seconds, &time->nanoseconds, FALSE);

	return data_size;
#elif defined(PTP_LAYER_2)
	return 0;
#else
	return socketSend(netPath->socketEvent, (void*)buf, length, time, &netPath->multicastAddr);
#endif
}

ssize_t netSendGeneral(NetPath *netPath, const octet_t *buf, int16_t  length)
{
#ifdef IDA_LWIP
	struct eth_hdr *ethhdr = (struct eth_hdr *)&general_tx_buffer[0];
	ethhdr->type = lwip_htons(ETHTYPE_PTP);
	SMEMCPY(&ethhdr->dest, &ptp_l2_dst_addr, ETH_HWADDR_LEN);
	SMEMCPY(&ethhdr->src,  ownMac.addr, ETH_HWADDR_LEN);

	memcpy(&general_tx_buffer[SIZEOF_ETH_HDR],buf,length);

	return ida_lwip_send_raw((void*)general_tx_buffer, length+SIZEOF_ETH_HDR, &txSem);
#elif defined(PTP_LAYER_2)
	return 0;
#else
	return socketSend(netPath->socketGeneral, (void*)buf, length, NULL, &netPath->multicastAddr);
#endif
}

ssize_t netSendPeerGeneral(NetPath *netPath, const octet_t *buf, int16_t  length)
{
#ifdef IDA_LWIP
	// TODO
	return 0;
#else
	return socketSend(netPath->socketGeneral, (void*)buf, length, NULL, &netPath->peerMulticastAddr);
#endif
}

ssize_t netSendPeerEvent(NetPath *netPath, const octet_t *buf, int16_t  length, TimeInternal* time)
{
#ifdef IDA_LWIP
	// TODO
	return 0;
#else
	return socketSend(netPath->socketEvent, (void*)buf, length, NULL, &netPath->peerMulticastAddr);
#endif
}

#if !defined(IDA_LWIP) && defined(PTP_LAYER_2)
int ptp_input(struct pbuf *p){
	if(ptp_rx_pbuf == (struct pbuf*)0){
		ptp_rx_pbuf = p;
		sys_sem_signal(&ptpRxSem);
		return 0;
	} else {
		return -1;
	}
}
#endif

